--[[ Socialize
     Maintain friend/ignore lists.

     Tries to help keep friend/ignore lists synchronized.
]]--

local addoninfo, soc = ...

local printf = Library.printf.printf
local printfh = Library.printf.printfh
local printfs = Library.printf.printfs
local printfhs = Library.printf.printfhs
local sprintf = Library.printf.sprintf
local sprintfh = Library.printf.sprintfh

local slower = string.lower
local supper = string.upper

local cron = Library.LibCron

local function keycount(a)
  local count = 0
  for k, v in pairs(a) do
    count = count + 1
  end
  return count
end

function soc.variables(event, addon)
  if addon == 'Socialize' then
    Socialize_SocialHistory = Socialize_SocialHistory or {}
    soc.data = Socialize_SocialHistory
    soc.data.known = soc.data.known or {}
    soc.data.me = soc.data.me or {}
  end
end

soc.incoming_events = {}
soc.events_to_process = {}
soc.setup_complete = false
soc.startup_happened = false
soc.tick_counter = 0
soc.stale = {}
soc.verbose = false
soc.ignored_early_startup = 0
soc.dirty = true

function soc.list_updated(iff, event, list)
  -- it turns out, updates before we are fully started up may not be
  -- informative.
  if not soc.startup_happened then
    return
  end
  -- let the system know we have started seeing events.
  -- let's just ignore events that come in early during startup.
  if soc.tick_counter < 1 then
    soc.ignored_early_startup = soc.ignored_early_startup + 1
    return
  end
  if soc.busy then
    return
  end
  if soc.ignored_early_startup > 0 then
    if soc.verbose then
      printf("%d message%s ignored during login.", soc.ignored_early_startup,
      	(soc.ignored_early_startup == 1) and "" or "s")
    end
    soc.ignored_early_startup = 0
  end
  local k = keycount(list)
  if k > 1 then
    if soc.verbose then
      printf("Ignoring probably-spurious %s event [%d items].", iff, k)
    end
    return
  end
  for k, v in pairs(list) do
    soc.incoming_events[#soc.incoming_events + 1] = { iff = iff, name = slower(k) }
  end
end

function soc.new_character()
  if soc.shard and soc.player then
    return false
  end
  if not soc.shard then
    local status, value = pcall(Inspect.Shard)
    if status then
      soc.shard = slower(value.name)
    end
  end
  if not soc.player then
    local status, value = pcall(Inspect.Unit.Detail, 'player')
    if status then
      soc.player = slower(value.name)
    end
  end
  if soc.shard and soc.player then
    soc.self = soc.player .. '@' .. soc.shard
    if soc.data.me[soc.self] then
      -- recheck every day or so.
      if type(soc.data.me[soc.self]) == 'number' then
        return soc.data.me[soc.self] + (60 * 60 * 24) < os.time()
      else
        return true
      end
    else
      return true
    end
  end
end

function soc.flip(iff)
  if iff == 'Friend' then
    return 'Ignore'
  else
    return 'Friend'
  end
end

function soc.canonicalize(source, dest)
  if type(source) == 'string' then
    local name = slower(source)
    local at = name:find('@')
    -- "@Unknown" appears to be an obselescent feature?
    if name:find('@unknown') then
      return nil
    end
    if not name:find('@') then
      name = name .. '@' .. soc.shard
    end
    return name
  end
  -- populate a table differently
  local found_any = false
  if type(source) == 'table' and type(dest) == 'table' then
    for k, v in pairs(source) do
      local fixed = soc.canonicalize(k)
      if fixed then
        dest[fixed] = v
        found_any = true
      end
    end
  end
  return found_any
end

function soc.pretty(name)
  local at = name:find('@')
  return supper(name:sub(1, 1)) .. name:sub(2, at) .. supper(name:sub(at + 1, at + 1)) .. name:sub(at + 2, -1)
end

-- import: add new entries from this character, but don't change
-- anything in this character's server-side data.
function soc.import_list(list, iff)
  local stamp = os.time()
  for k, v in pairs(list) do
    local name = soc.canonicalize(k)
    printf("Considering %s (%s)", name, tostring(soc.data.known[name]))
    if not soc.data.known[name] then
      printf("New %s: %s", iff, soc.pretty(name))
      soc.data.known[name] = {
        from = soc.self,
	note = v.note,
	iff = iff,
	stamp = stamp,
	stale = false,
      }
    else
      if v.note and not soc.data.known[name].note then
        printf("%s has a note: %s", soc.pretty(name), v.note)
      end
    end
  end
end

function soc.current_data()
  local basedata = {
    Friend = Inspect.Social.Friend.Detail(Inspect.Social.Friend.List()),
    Ignore = Inspect.Social.Ignore.Detail(Inspect.Social.Ignore.List())
  }
  local cfdata = {
    Friend = {},
    Neutral = {},
    Ignore = {},
  }
  -- missing data?
  local found_any = false
  if soc.canonicalize(basedata.Friend, cfdata.Friend) then
    found_any = true
  end
  if soc.canonicalize(basedata.Ignore, cfdata.Ignore) then
    found_any = true
  end

  cfdata.friends = keycount(cfdata.Friend)
  cfdata.ignores = keycount(cfdata.Ignore)
  return found_any and cfdata
end

function soc.sync_new_character()
  local cfdata = soc.current_data()
  if not cfdata then
    printf("Can't import friend and ignore lists yet, they're not loaded.")
    return
  end
  -- first, we import anything known to them but not known to us.
  soc.import_list(cfdata.Friend, 'Friend')
  soc.import_list(cfdata.Ignore, 'Ignore')
  -- and if we did that, we're probably now "known".
  printf("Imported friend and ignore lists.")
  soc.data.me[soc.self] = os.time()
end

function soc.setup(event)
  if soc.startup_happened then
    return
  end
  if not soc.shard then
    local status, value = pcall(Inspect.Shard)
    if status then
      soc.shard = slower(value.name)
    end
  end
  soc.startup_happened = true
  soc.fire_when_ready()
end

function soc.fire_when_ready()
  if soc.setup_complete then
    return
  end
  if soc.verbose then
    printf("Socialize coming online...")
  end
  if soc.new_character() then
    if soc.data.me[soc.self] then
      printf("Rechecking/importing local friend/ignore lists in case of changes.")
    else
      printf("New character. Doing initial setup.")
    end
    soc.sync_new_character()
  end
  soc.crontask = cron.new('Socialize', 5, true, false, soc.process_events)
  soc.setup_complete = true
end

soc.old_event_count = 0

function soc.process_events()
  -- got a new event? Okay, let's wait a tick for the lists to settle.
  soc.tick_counter = soc.tick_counter + 1
  if #soc.incoming_events > soc.old_event_count then
    soc.old_event_count = #soc.incoming_events
    return
  end
  local stamp = os.time()
  local cfdata = soc.current_data()
  if not cfdata then
    -- we'll have to try again when we have data available?
    return
  end
  local events = soc.incoming_events
  soc.incoming_events = {}
  for i = 1, #events do
    local e = events[i]
    local name = soc.canonicalize(e.name)

    -- something happened, so we might care.
    soc.dirty = true

    -- if we have an existing entry, this may be a change...
    if name then
      if not soc.data.known[name] then
	if cfdata[e.iff][name] then
	  -- newly added!
	  soc.data.known[name] = {
	    from = soc.self,
	    note = cfdata[e.iff][name].note,
	    iff = e.iff,
	    stale = false,
	    stamp = stamp,
	  }
	else
	  -- probably a removal, but I guess we don't care?
	  printf("Apparent removal for unknown %s from %s.", soc.pretty(name), e.iff)
	end
      else
	if cfdata[e.iff][name] then
	  -- looks like an addition
	  if soc.data.known[name].iff ~= e.iff then
	    printf("Looks like %s changed from %s to %s.", soc.pretty(name),
	     soc.data.known[name].iff, e.iff)
	    soc.data.known[name].iff = e.iff
	  end
	  if cfdata[e.iff][name].note then
	    if soc.data.known[name].note ~= cfdata[e.iff][name].note then
	      printf("New note for %s: %s", soc.pretty(name), cfdata[e.iff][name].note)
	      if (soc.data.known[name].note) then
	        printf("Old note was: %s", soc.data.known[name].note)
	      end
	      soc.data.known[name].note = cfdata[e.iff][name].note
	    end
	  end
	else
	  -- looks like a removal
	  if soc.data.known[name].iff == e.iff then
	    printf("Looks like %s was removed from %s.", soc.pretty(name), e.iff)
	    soc.data.known[name].iff = 'Neutral'
	  else
	    printf("Got an apparent %s removal for %s, but no effect?",
	      e.iff, soc.pretty(name))
	  end
	end
      end
    end
  end
  -- and now we may have changes to apply.
  if soc.dirty then
    soc.apply_settings()
  end
end

function soc.add(iff, name, note)
  -- just in case
  pcall(Command.Social[soc.flip(iff)].Remove, name)
  -- we might get an event from this, which could show up slightly before
  -- the list gets updated (?) making us confused.
  -- soc.skip_event[name] = true
  -- then we add them to the right list.
  pcall(Command.Social[iff].Add, name)
  pcall(Command.Social[iff].Note, name, note)
  -- remember that we tried to do this.
  soc.stale[name] = iff
end

function soc.remove(iff, name)
  pcall(Command.Social[iff].Remove, name)
  soc.stale[name] = 'Neutral'
end

function soc.apply_settings()
  soc.busy = true
  local stamp = os.time()
  local cfdata = soc.current_data()
  local added_items = false
  local added_notes = false
  if soc.verbose then
    printf("Apply settings:")
  end
  if not cfdata then
    if soc.verbose then
      printf("No data yet?")
    end
    return
  else
    if soc.verbose then
      printf("Existing: Friend %d, Ignore %d.", cfdata.friends, cfdata.ignores)
    end
  end
  for name, v in pairs(soc.data.known) do
    -- no entry at all?
    if v.iff == 'Neutral' then
      if not v.stale then
	if soc.stale[name] == v.iff then
	  printf("Apparently can't remove %s entry for %s. Marking as stale.", v.iff, soc.pretty(name))
	  v.stale = true
	else
          if cfdata.Friend[name] then
            printf("Removing Friend entry for %s.", soc.pretty(name))
            soc.remove('Friend', name)
          elseif cfdata.Ignore[name] then
            printf("Removing Ignore entry for %s.", soc.pretty(name))
            soc.remove('Ignore', name)
	  else
	    -- If you aren't on either list, this is probably a bogus entry.
	    v.stale = true
          end
        end
      end
    else
      if not cfdata[v.iff] then
        printf("Help! For %s, iff is %s.", name, v.iff)
      end
      -- printf("%s: %s. [%s]", soc.pretty(name), v.iff, cfdata[v.iff][name] and "existing" or "new")
      if cfdata[v.iff][name] then
	-- well, maybe it's NOT stale, then?
	v.stale = false
        if v.note and not cfdata[v.iff][name].note then
          pcall(Command.Social[v.iff].Note, name, v.note)
	  added_notes = true
        end
      else
	if not v.stale then
	  if soc.stale[name] == v.iff then
	    printf("Apparently can't add %s entry for %s. Marking as stale.", v.iff, soc.pretty(name))
	    v.stale = true
	  else
            printf("Adding %s entry for %s.", v.iff, soc.pretty(name))
            soc.add(v.iff, name, v.note)
	  end
        end
      end
    end
  end
  if added_notes then
    printf("Updated friend/ignore notes.")
  end
  soc.busy = false
  soc.dirty = false
end

function soc.slashcommand(args)
  if args.v then
    soc.verbose = not soc.verbose
    printf("Verbose mode: %s", tostring(soc.verbose))
  end
  if args.r then
    printf("Resetting stale markers.")
    for k, v in pairs(soc.data.known) do
      v.stale = false
    end
  end
  if args.u then
    printf("Running an event processing loop.")
    soc.process_events()
  end
  for i = 1, #args.leftover_args do
    local name = soc.canonicalize(args.leftover_args[i])
    local pretty = soc.pretty(name)
    local data = soc.data.known[name]
    if data then
      if data.note then
        printf("%s: %s, added by %s. Note: %s.",
          name, data.iff, data.from, data.note)
      else
        printf("%s: %s, added by %s. No note.",
          name, data.iff, data.from)
      end
    else
      printf("Unknown: %s.", pretty)
    end
  end
end

Library.LibGetOpt.makeslash("ru", "Socialize", "soc", soc.slashcommand)

Command.Event.Attach(Event.Addon.SavedVariables.Load.End, soc.variables, "variable check")
Command.Event.Attach(Event.Addon.Startup.End, soc.setup, "initial setup")
Command.Event.Attach(Event.Social.Friend, function(event, list) soc.list_updated('Friend', event, list) end, "friend list update handler")
Command.Event.Attach(Event.Social.Ignore, function(event, list) soc.list_updated('Ignore', event, list) end, "ignore list update handler")
